# -*- coding: utf-8 -*-
"""Credential fetcher for 2nd Screen app."""
import logging
import socket

from .errors import CredentialTimeout, UnknownDDPResponse

_LOGGER = logging.getLogger(__name__)

DEFAULT_DEVICE_NAME = 'pyps4-2ndScreen'
STANDBY = '620 Server Standby'
HOST_ID = '1234567890AB'
UDP_IP = '0.0.0.0'
REQ_PORT = 997
DDP_PORT = 987
DDP_VERSION = '00020020'

"""
PS4 listens on ports 987 and 997 (Priveleged).
Must run command on python path:
"sudo setcap 'cap_net_bind_service=+ep' /usr/bin/python3.5"
"""


class Credentials:
    """The PS4 Credentials object. Masquerades as a PS4 to get credentials."""

    def __init__(self, device_name=DEFAULT_DEVICE_NAME):
        """Init Cred Server."""
        self.sock = None
        self.response = {
            'host-id': HOST_ID,
            'host-type': 'PS4',
            'host-name': device_name,
            'host-request-port': REQ_PORT
        }
        self.start()

    def start(self):
        """Start Cred Server."""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock = sock
        except socket.error:
            _LOGGER.error("Failed to create socket")
            return
        sock.settimeout(3)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind((UDP_IP, DDP_PORT))
        except socket.error as error:
            _LOGGER.error(
                "Could not bind to port %s; \
                Ensure port is accessible and unused, %s",
                DDP_PORT, error)
            return

    def listen(self, timeout=120):
        """Listen and respond to requests."""
        self.sock.settimeout(timeout)
        data = None
        address = None
        response = None
        _LOGGER.info(
            "Starting Credential Service with Timeout of %s seconds.",
            timeout)
        while 1:
            try:
                response = self.sock.recvfrom(1024)
            except socket.error:
                self.sock.close()
            if not response:
                _LOGGER.info(
                    "Credential service has timed out with no response.")
                raise CredentialTimeout
            data = response[0]
            address = response[1]
            if parse_ddp_response(data, 'search') == 'search':
                _LOGGER.debug("Search from: %s", address)
                msg = get_ddp_message(STANDBY, self.response)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                try:
                    self.sock.sendto(msg.encode('utf-8'), address)
                except socket.error:
                    self.sock.close()
            if parse_ddp_response(data, 'wakeup') == 'wakeup':
                _LOGGER.debug("Wakeup from: %s", address)
                creds = get_creds(data)
                self.sock.close()
                return creds
        return None


def get_ddp_message(status, data=None):
    """Get DDP message."""
    msg = u'HTTP/1.1 {}\n'.format(status)
    if data is not None:
        for key, value in data.items():
            msg += u'{}:{}\n'.format(key, value)
    msg += u'device-discovery-protocol-version:{}\n'.format(DDP_VERSION)
    return msg


def parse_ddp_response(response, listen_type):
    """Parse the response."""
    rsp = response.decode('utf-8')
    if listen_type == 'search':
        if 'SRCH' in rsp:
            return 'search'
    elif listen_type == 'wakeup':
        if 'WAKEUP' in rsp:
            return 'wakeup'
    else:
        raise UnknownDDPResponse
    return None


def get_creds(response):
    """Return creds."""
    keys = {}
    data = response.decode('utf-8')
    for line in data.splitlines():
        line = line.strip()
        if ":" in line:
            value = line.split(':')
            keys[value[0]] = value[1]
    cred = keys['user-credential']
    return cred



# -*- coding: utf-8 -*-
"""Methods for PS4 Object."""
from __future__ import print_function
import json
import logging
import time
import asyncio

import aiohttp

from .connection import Connection, AsyncConnection
from .ddp import (get_status, launch, wakeup,
                  get_ddp_launch_message, get_ddp_wake_message)
from .errors import (NotReady, PSDataIncomplete,
                     UnknownButton, LoginFailed)
from .media_art import (async_get_ps_store_requests,
                        get_lang, parse_data, COUNTRIES,
                        async_prepare_tumbler)

_LOGGER = logging.getLogger(__name__)

BUTTONS = {'up': 1,
           'down': 2,
           'right': 4,
           'left': 8,
           'enter': 16,
           'back': 32,
           'option': 64,
           'ps': 128,
           'key_off': 256,
           'cancel': 512,
           'open_rc': 1024,
           'close_rc': 2048}

STATUS_OK = 200
STATUS_STANDBY = 620

DEFAULT_DEVICE_NAME = 'pyps4-2ndScreen'


def opencredential_file(filename):
    """Open credential file."""
    return json.load(open(filename))


def delay(seconds):
    """Delay in seconds."""
    start_time = time.time()
    while time.time() - start_time < seconds:
        pass


class Ps4():
    """The PS4 object."""

    def __init__(self, host, credential=None, credentials_file=None,
                 device_name=DEFAULT_DEVICE_NAME):
        """Initialize the instance.

        Keyword arguments:
            host -- the host IP address
            credential -- the credential string
            credential_file -- credential file in working directory
        """
        self.host = host
        self.credential = None
        self.device_name = device_name
        self._socket = None
        self._power_on = False
        self._power_off = False
        self.msg_sending = False
        self.status = None
        self.connected = False
        self.client = None
        self.ps_cover = None
        self.ps_name = None
        self.loggedin = False

        if credential:
            self.credential = credential
        if credentials_file:
            creds = opencredential_file(credentials_file)
            self.credential = creds['user-credential']

        self.connection = Connection(self, credential=self.credential)

        if self.client is not None:
            self.client.add_ps4(ps4=self)

    def _prepare_connection(self):
        self.launch()
        delay(0.5)
        _LOGGER.debug("Connection prepared")

    def open(self):
        """Open a connection to the PS4."""
        if self.status is None:
            self.get_status()
        if not self.is_running:
            raise NotReady("PS4 is not On")

        self._prepare_connection()
        if self.connected is False:
            self.connection.connect()
            login = self.connection.login()
            if login is True:
                self.connected = True
                return True
            return False
        return True

    def close(self):
        """Close the connection to the PS4."""
        self.connection.disconnect()
        self.connected = False
        _LOGGER.debug("Disconnecting from PS4 @ %s", self.host)
        return True

    def get_status(self) -> dict:
        """Get current status info."""
        import socket

        try:
            self.status = get_status(self.host)
        except socket.timeout:
            _LOGGER.debug("PS4 @ %s status timed out", self.host)
            self.status = None
            return self.status
        else:
            if self.status is not None:
                if self.is_standby:
                    self.connected = False
                    self.loggedin = False
                return self.status
            return None

    def launch(self):
        """Launch."""
        launch(self.host, self.credential)

    def wakeup(self):
        """Wakeup."""
        wakeup(self.host, self.credential)
        self._power_on = True

    def login(self, pin=None):
        """Login."""
        self.open()
        is_login = self.connection.login(pin)
        if is_login is False:
            raise LoginFailed("PS4 Refused Connection")
        self.close()
        return is_login

    def standby(self, retry=2):
        """Standby."""
        retries = 0
        while retries < retry:
            self.open()
            if self.connection.standby():
                self.close()
                return True
            self.close()
            retries += 1
        return False

    def start_title(self, title_id, running_id=None, retry=2):
        """Start title.

        `title_id`: title to start
        'running_id': Title currently running,
        Use to confirm closing of current title.
        """
        if self.msg_sending:
            _LOGGER.warning("PS4 already sending message.")
            return False
        retries = 0
        while retries < retry:
            self.msg_sending = True
            if self.open():
                if self.connection.start_title(title_id):

                    # Auto confirm prompt to close title.
                    if running_id is not None:
                        delay(1)
                        self.remote_control('enter')
                    elif running_id == title_id:
                        _LOGGER.warning("Title: %s already started", title_id)
                    self.msg_sending = False
                    return True
                else:
                    self.close()
                    retries += 1
                    delay(1)
        self.msg_sending = False
        return False

    def remote_control(self, button_name, hold_time=0):
        """Send a remote control button press.

        Documentation from ps4-waker source:
        near as I can tell, here's how this works:
         - For a simple tap, you send the key with holdTime=0,
           followed by KEY_OFF and holdTime = 0
         - For a long press/hold, you still send the key with
           holdTime=0, the follow it with the key again, but
           specifying holdTime as the hold duration.
         - After sending a direction, you should send KEY_OFF
           to clean it up (since it can just be held forever).
           Doing this after a long-press of PS just breaks it,
           however.
        """
        if self.msg_sending:
            _LOGGER.warning("PS4 already sending message.")
            return
        self.msg_sending = True
        button_name = button_name.lower()
        if button_name not in BUTTONS.keys():
            raise UnknownButton("Button: {} is not valid".format(button_name))
        operation = BUTTONS[button_name]
        if self.open():
            _LOGGER.debug("Sending RC Command: %s", button_name)
            if not self.connection.remote_control(operation, hold_time):
                self.close()
        self.msg_sending = False

    def send_status(self):
        """Send connection status to PS4."""
        if self.connected is True:
            is_loggedin = self.connection.send_status()
            if is_loggedin is False:
                self.close()

    async def async_get_ps_store_data(self, title, title_id, region):
        """Get and Parse Responses."""
        lang = get_lang(region)
        _LOGGER.debug("Starting search request")
        async with aiohttp.ClientSession() as session:
            responses = await async_get_ps_store_requests(
                title, region, session)
            for response in responses:
                try:
                    result_item = parse_data(response, title_id, lang)
                except (TypeError, AttributeError):
                    result_item = None
                    raise PSDataIncomplete
                if result_item is not None:
                    break

            if result_item is None:
                try:
                    result_item = await async_prepare_tumbler(
                        title, title_id, region, session)
                except (TypeError, AttributeError):
                    result_item = None
                    raise PSDataIncomplete

            await session.close()

            if result_item is not None:
                _LOGGER.debug("Found Title: %s, URL: %s",
                              result_item.name, result_item.cover_art)
                self.ps_name = result_item.name
                self.ps_cover = result_item.cover_art
                return result_item

            return None

    async def async_search_all_ps_data(self, title, title_id, timeout=10):
        """Search for title in all regions."""
        _LOGGER.debug("Searching all databases...")
        tasks = []
        for region in COUNTRIES:
            search_func = self.async_get_ps_store_data(title, title_id, region)
            task = asyncio.ensure_future(
                self._async_search_region(search_func))
            tasks.append(task)

        done, pending = await asyncio.wait(
            tasks, timeout=timeout, return_when=asyncio.FIRST_COMPLETED)

        # Return First Result.
        data = None
        for completed in done:
            try:
                data = completed.result()
            except asyncio.InvalidStateError:
                data = None
            if data is not None:
                for task in pending:
                    task.cancel()
                return data
        return None

    async def _async_search_region(self, task):
        result_item = await task
        if result_item is not None:
            return result_item
        return None

    @property
    def is_running(self):
        """Return True if the PS4 is running."""
        if self.status is not None:
            if self.status['status_code'] == STATUS_OK:
                return True
        return False

    @property
    def is_standby(self):
        """Return True if the PS4 is in standby."""
        if self.status is not None:
            if self.status['status_code'] == STATUS_STANDBY:
                return True
        return False

    @property
    def is_available(self):
        """Return if the PS4 is available."""
        if self.status is not None:
            return True
        return False

    @property
    def system_version(self):
        """Get the system version."""
        if self.status is not None:
            return self.status['system-version']
        return None

    @property
    def host_id(self):
        """Get the host id."""
        if self.status is not None:
            return self.status['host-id']
        return None

    @property
    def host_name(self):
        """Get the host name."""
        if self.status is not None:
            return self.status['host-name']
        return None

    @property
    def running_app_titleid(self):
        """Return the title Id of the running application."""
        if self.status is not None:
            if 'running-app-titleid' in self.status:
                return self.status['running-app-titleid']
        return None

    @property
    def running_app_name(self):
        """Return the name of the running application."""
        if self.status is not None:
            if 'running-app-name' in self.status:
                return self.status['running-app-name']
        return None

    @property
    def running_app_ps_cover(self):
        """Return the URL for the title cover art."""
        if self.running_app_titleid is None:
            self.ps_cover = None
        return self.ps_cover

    @property
    def running_app_ps_name(self):
        """Return the name fetched from PS Store."""
        if self.running_app_titleid is None:
            self.ps_name = None
        return self.ps_name


class Ps4Async(Ps4):
    """Async Version of Ps4 Class."""

    def __init__(self, host, credential=None, credentials_file=None,
                 device_name=None):
        """Inherit Class."""
        super().__init__(host, credential, credentials_file,
                         device_name)
        self.ddp_protocol = None
        self.tcp_transport = None
        self.tcp_protocol = None
        self.task_queue = None
        self.poll_count = 0
        self.unreachable = False

        self.connection = AsyncConnection(self, self.credential)

    def open(self):
        """Not Implemented."""
        raise NotImplementedError

    def get_status(self) -> dict:
        """Get current status info."""
        if self.ddp_protocol is not None:
            self.ddp_protocol.send_msg(self)
            if self.status is not None:
                if self.is_standby:
                    self.connected = False
                    self.loggedin = False

                    # Ensure that connection is closed.
                    if self.tcp_protocol is not None:
                        asyncio.ensure_future(self.close())

                return self.status
            return None

        else:
            return super().get_status()

    def launch(self):
        """Launch."""
        if self.ddp_protocol is None:
            _LOGGER.error("DDP Protocol does not exist/Not ready")
        else:
            self.ddp_protocol.send_msg(
                self, get_ddp_launch_message(self.credential))

    def wakeup(self):
        """Wakeup."""
        if self.ddp_protocol is None:
            _LOGGER.error("DDP Protocol does not exist")
        else:
            self._power_on = True
            self._power_off = False
            self.ddp_protocol.send_msg(
                self, get_ddp_wake_message(self.credential))

    async def login(self, pin=None):
        """Login."""
        if self.tcp_protocol is None:
            _LOGGER.info("TCP Protocol does not exist")
        else:
            await self.tcp_protocol.login(pin)

    async def standby(self, retry=None):
        """Standby."""
        if retry is not None:
            _LOGGER.info("Retries not implemented")
        if self.tcp_protocol is None:
            _LOGGER.info("TCP Protocol does not exist")
        else:
            await self.tcp_protocol.standby()
            self._power_off = True

    async def start_title(self, title_id, running_id=None, retry=None):
        """Start title."""
        if running_id is None:
            if self.running_app_titleid is not None:
                running_id = self.running_app_titleid

        if retry is not None:
            _LOGGER.info("Retries not implemented")
        if self.tcp_protocol is None:
            _LOGGER.info("TCP Protocol does not exist")

            # Queue task upon login.
            task = ('start_title', title_id, running_id)
            self.task_queue = task
            self.wakeup()

        else:
            await self.tcp_protocol.start_title(title_id, running_id)

    async def remote_control(self, button_name, hold_time=0):
        """Remote Control."""
        button_name = button_name.lower()
        if button_name not in BUTTONS.keys():
            raise UnknownButton(
                "Button: {} is not valid".format(button_name))
        operation = BUTTONS[button_name]

        if self.tcp_protocol is None:
            _LOGGER.info("TCP Protocol does not exist")

            # Queue task upon login.
            task = ('remote_control', operation, hold_time)
            self.task_queue = task
            self.wakeup()

        else:
            await self.tcp_protocol.remote_control(operation, hold_time)

    async def close(self):
        """Close Transport."""
        self._close()

    def _close(self):
        if self.tcp_protocol is None:
            _LOGGER.info("TCP Protocol @ %s already disconnected", self.host)
        else:
            self.tcp_protocol.disconnect()
            self.tcp_transport = None
            self.tcp_protocol = None

    async def async_connect(self, auto_login=True):
        """Connect."""
        if not self.connected:
            if self.status is None:
                self.get_status()
            if not self.is_available:
                raise NotReady(
                    "PS4 is not available or powered off. Check connection.")
            if not self._power_off:
                if self.is_standby:
                    raise NotReady("PS4 is not On")
                try:
                    self._prepare_connection()
                    self.tcp_transport, self.tcp_protocol =\
                        await self.connection.async_connect(self)
                except (OSError, ConnectionRefusedError):
                    _LOGGER.info("PS4 Refused Connection")
                    self.connected = False
                    self.loggedin = False
                else:
                    self.connected = True
                    if self._power_on:
                        if auto_login:
                            await self.login()
                    self._power_on = False



# -*- coding: utf-8 -*-
"""Client for PS4."""
import logging
import threading
import time
import queue

_LOGGER = logging.getLogger(__name__)

DEFAULT_INTERVAL = 3
DEFAULT_INTERVAL_KEEP_ALIVE = 30


def _should_poll(last, interval):
    """Poll timer."""
    time_since = time.time() - last
    if time_since > interval:
        return True
    return False


class PS4Client():
    """Client for handling background IO."""

    def __init__(self):
        """Init."""
        self.listeners = {}
        self.started = []

    def start_all(self):
        """Start all listeners."""
        if self.listeners:
            for listener in self.listeners:
                listener.start()

    def stop_all(self):
        """Stop all Listeners."""
        if self.listeners:
            for listener in self.listeners:
                listener.stop.set()

    def remove_all(self):
        """Remove all listeners."""
        self.stop_all()
        self.listeners = {}

    def start_listener(self, ps4):
        """Start listener thread."""
        _LOGGER.debug("Starting listener @ %s", ps4._host)  # noqa: pylint: disable=protected-access
        self.started.append(self.listeners[ps4])
        self.listeners[ps4].start()

    def stop_listener(self, ps4):
        """Stop a Listener."""
        if ps4 in self.listeners:
            self.listeners[ps4].stop.set()
            self.started.remove(self.listeners[ps4])

    def remove_listener(self, ps4):
        """Stop and Remove a Listener."""
        self.stop_listener(ps4)
        self.listeners.pop(ps4)

    def schedule_task(self, ps4, task, *args):
        """Schedule socket task in thread."""
        active = self.get_active()
        if active is not None:
            if active.ps4 != ps4:
                active.ps4.close()
        if not ps4.msg_sending:
            listener = self.listeners[ps4]
            if args:
                task = (task, *args)
            listener.queue.put(task)
        else:
            _LOGGER.info("PS4 already has a task in progress")

    def add_ps4(self, ps4):
        """Add PS4 to listeners."""
        if ps4 not in self.listeners:
            self.listeners[ps4] = Listener(client=self, ps4=ps4)
        else:
            _LOGGER.warning("Listener already exists")

    def add_callback(self, ps4, callback):
        """Add Callback for listener."""
        assert callable(callback) is True, "Callback must be callable"
        if ps4 in self.listeners:
            self.listeners[ps4].callbacks.append(callback)

    def get_active(self):
        """Return logged in PS4."""
        for listener in self.listeners.values():
            if listener.ps4.connected:
                return listener
        return None


class Listener(threading.Thread):
    """Listener per PS4 to handle sockets."""

    def __init__(self, client, ps4):
        """Init."""
        super().__init__()
        self.client = client
        self.ps4 = ps4
        self.interval = DEFAULT_INTERVAL
        self.interval_keep_alive = DEFAULT_INTERVAL_KEEP_ALIVE
        self.last_poll = 0
        self.last_keep_alive = 0
        self.callbacks = []
        self.running = False
        self._status = None
        self.queue = queue.Queue()

        self.stop = threading.Event()
        self.block = threading.Event()

    def run(self):
        """Run Listener."""
        while not self.stop.is_set():
            self.running = True
            while not self.queue.empty():
                if not self.block.is_set():
                    # Block queue.
                    self.block.set()
                    item = self.queue.get()

                    # If args are passed in.
                    if isinstance(item, tuple):
                        args = item[1:]
                        item = item[0]
                        result = item(*args)
                    else:
                        result = item()

                    if item == self.ps4.get_status:
                        self._check_status(result)
                        self.last_poll = time.time()
                    elif item == self.ps4.send_status:
                        self.last_keep_alive = time.time()

                    # Reset intervals for status.
                    else:
                        self.last_poll = time.time()
                        self.last_keep_alive = time.time()

                    # Unblock queue.
                    self.block.clear()

            # Call when queue is empty.
            # Queue status update.
            if _should_poll(self.last_poll, self.interval):
                self.queue.put(self.ps4.get_status, False)

            # Queue Keep Alive.
            if self.ps4.connected:
                if _should_poll(self.last_keep_alive,
                                self.interval_keep_alive):
                    self.queue.put(self.ps4.send_status, False)

        # If stopping.
        self.queue.join()
        self.running = False
        _LOGGER.debug("Stopping Listener @ %s", self.ps4._host)   # noqa: pylint: disable=protected-access

    def _notify_callbacks(self):
        """Call all callbacks."""
        if self.callbacks:
            for callback in self.callbacks:
                callback()

    def _check_status(self, status):
        if status != self._status:
            _LOGGER.debug(
                "PS4 @ %s: Status changed to %s", self.ps4._host, status)  # noqa: pylint: disable=protected-access
            self._status = status
            self._notify_callbacks()



"""Helpers."""
import logging

from .errors import NotReady, LoginFailed

_LOGGER = logging.getLogger(__name__)


class Helper:
    """Helpers for PS4."""

    def __init__(self):
        """Init Class."""

    def has_devices(self, host=None):  # noqa: pylint: disable=no-self-use
        """Return if there are devices that can be discovered."""
        from .ddp import Discovery

        _LOGGER.debug("Searching for PS4 Devices")
        discover = Discovery()
        devices = discover.search(host)
        for device in devices:
            _LOGGER.debug("Found PS4 at: %s", device['host-ip'])
        return devices

    def link(self, host, creds, pin, device_name=None):  # noqa: pylint: disable=no-self-use
        """Perform pairing with PS4."""
        from .ps4 import Ps4

        ps4 = Ps4(host, creds, device_name=device_name)
        is_ready = True
        is_login = True
        try:
            ps4.login(pin)
        except NotReady:
            is_ready = False
        except LoginFailed:
            is_login = False
        return is_ready, is_login

    def get_creds(self, device_name=None):  # noqa: pylint: disable=no-self-use
        """Return Credentials."""
        from .credential import Credentials

        credentials = Credentials(device_name)
        return credentials.listen()

    def port_bind(self, ports):  # noqa: pylint: disable=no-self-use
        """Try binding to ports."""
        import socket

        for port in ports:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(1)
                sock.bind(('0.0.0.0', port))
                sock.close()
            except socket.error:
                sock.close()
                return int(port)
            return None



# -*- coding: utf-8 -*-
"""Init File for pyps4_homeassistant."""
from .ps4 import *
from .connection import *
from .credential import *
from .helpers import *
from .ddp import *
from .media_art import *
from .errors import *
from .client import *



# -*- coding: utf-8 -*-
"""TCP Handling for PS4."""
from __future__ import print_function

import binascii
import logging
import socket
import time
import asyncio
from typing import cast

from construct import (Bytes, Const, Int32ul, Padding, Struct)
from Cryptodome.Cipher import AES, PKCS1_OAEP
from Cryptodome.PublicKey import RSA

_LOGGER = logging.getLogger(__name__)

TCP_PORT = 997
STATUS_REQUEST = b'0c000000120000000000000000000000'
RANDOM_SEED = \
    b'\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
TIMEOUT = 5

PUBLIC_KEY = (
    '-----BEGIN PUBLIC KEY-----\n'
    'MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAxfAO/MDk5ovZpp7xlG9J\n'
    'JKc4Sg4ztAz+BbOt6Gbhub02tF9bryklpTIyzM0v817pwQ3TCoigpxEcWdTykhDL\n'
    'cGhAbcp6E7Xh8aHEsqgtQ/c+wY1zIl3fU//uddlB1XuipXthDv6emXsyyU/tJWqc\n'
    'zy9HCJncLJeYo7MJvf2TE9nnlVm1x4flmD0k1zrvb3MONqoZbKb/TQVuVhBv7SM+\n'
    'U5PSi3diXIx1Nnj4vQ8clRNUJ5X1tT9XfVmKQS1J513XNZ0uYHYRDzQYujpLWucu\n'
    'ob7v50wCpUm3iKP1fYCixMP6xFm0jPYz1YQaMV35VkYwc40qgk3av0PDS+1G0dCm\n'
    'swIDAQAB\n'
    '-----END PUBLIC KEY-----')


def delay(seconds):
    """Delay in seconds."""
    start_time = time.time()
    while time.time() - start_time < seconds:
        pass


def _get_public_key_rsa():
    """Get RSA Key."""
    key = RSA.importKey(PUBLIC_KEY)
    return key.publickey()


def _handle_response(command, msg):
    """Return Pass/Fail for sent message."""
    pass_response = {
        'send_status': [18],
        'remote_control': [18],
        'start_title': [11, 18],
        'standby': [27],
        'login': [0, 17]
    }
    if command is not None:
        _LOGGER.debug("Handling command: %s", command)
        if command == 'login':
            response_byte = msg[8]
            if response_byte in pass_response['login']:
                _LOGGER.debug("Login Successful")
                return True
            _LOGGER.debug("Login Failed")
            return False

        response_byte = msg[4]
        _LOGGER.debug("RECV: %s for Command: %s", response_byte, command)
        if response_byte not in pass_response[command]:
            _LOGGER.warning("Command: %s Failed", command)
            return False
    return True


def _get_hello_request():
    fmt = Struct(
        'length' / Const(b'\x1c\x00\x00\x00'),
        'type' / Const(b'\x70\x63\x63\x6f'),
        'version' / Const(b'\x00\x00\x02\x00'),
        'dummy' / Padding(16),
    )

    msg = fmt.build({})
    return msg


def _parse_hello_request(msg):
    fmt = Struct(
        'length' / Int32ul,
        'type' / Int32ul,
        'version' / Int32ul,
        'dummy' / Bytes(8),
        'seed' / Bytes(16),
    )

    data = fmt.parse(msg)
    return data


def _get_handshake_request(seed):
    fmt = Struct(
        'length' / Const(b'\x18\x01\x00\x00'),
        'type' / Const(b'\x20\x00\x00\x00'),
        'key' / Bytes(256),
        'seed' / Bytes(16),
    )

    recipient_key = _get_public_key_rsa()
    cipher_rsa = PKCS1_OAEP.new(recipient_key)
    key = cipher_rsa.encrypt(RANDOM_SEED)

    _LOGGER.debug('key %s', binascii.hexlify(key))

    msg = fmt.build({'key': key, 'seed': seed})
    return msg


def _get_login_request(credential, name=None, pin=None):
    fmt = Struct(
        'length' / Const(b'\x80\x01\x00\x00'),
        'type' / Const(b'\x1e\x00\x00\x00'),
        'pass_code' / Const(b'\x00\x00\x00\x00'),
        'magic_number' / Const(b'\x01\x02\x00\x00'),
        'account_id' / Bytes(64),
        'app_label' / Bytes(256),
        'os_version' / Bytes(16),
        'model' / Bytes(16),
        'pin_code' / Bytes(16),
    )

    if pin is None:
        pin = b''
    else:
        pin = pin.encode()

    name = name.encode()

    config = {
        'app_label': b'PlayStation'.ljust(256, b'\x00'),
        'account_id': credential.encode().ljust(64, b'\x00'),
        'os_version': b'4.4'.ljust(16, b'\x00'),
        'model': name.ljust(16, b'\x00'),
        'pin_code': pin.ljust(16, b'\x00'),
    }

    _LOGGER.debug('config %s', config)

    msg = fmt.build(config)
    return msg


def _get_standby_request():
    fmt = Struct(
        'length' / Const(b'\x08\x00\x00\x00'),
        'type' / Const(b'\x1a\x00\x00\x00'),
        'dummy' / Padding(8),
    )

    msg = fmt.build({})
    return msg


def _get_boot_request(title_id):
    fmt = Struct(
        'length' / Const(b'\x18\x00\x00\x00'),
        'type' / Const(b'\x0a\x00\x00\x00'),
        'title_id' / Bytes(16),
        'dummy' / Padding(8),
    )

    msg = fmt.build({'title_id': title_id.encode().ljust(16, b'\x00')})
    return msg


def _get_remote_control_request(operation, hold_time) -> list:
    fmt = Struct(
        'length' / Const(b'\x10\x00\x00\x00'),
        'type' / Const(b'\x1c\x00\x00\x00'),
        'op' / Int32ul,
        'hold_time' / Int32ul,
    )
    # Prebuild required remote messages."""
    msg = []
    if operation != 128:
        msg.append(fmt.build({'op': 1024, 'hold_time': 0}))  # Open RC
        msg.append(fmt.build({'op': operation, 'hold_time': hold_time}))
        msg.append(fmt.build({'op': 256, 'hold_time': 0}))  # Key Off
        msg.append(fmt.build({'op': 2048, 'hold_time': 0}))  # Close RC
    else:  # PS
        msg.append(fmt.build({'op': 1024, 'hold_time': 0}))  # Open RC
        msg.append(fmt.build({'op': operation, 'hold_time': hold_time}))
        msg.append(fmt.build({'op': operation, 'hold_time': 1}))
        msg.append(fmt.build({'op': 256, 'hold_time': 0}))  # Key Off

    return msg


def _get_remote_control_close_request():
    fmt = Struct(
        'length' / Const(b'\x10\x00\x00\x00'),
        'type' / Const(b'\x1c\x00\x00\x00'),
        'op' / Int32ul,
        'hold_time' / Int32ul,
    )

    msg = fmt.build({'op': 2048, 'hold_time': 0})  # Close RC
    return msg


def _get_status_ack():
    fmt = Struct(
        'length' / Const(b'\x0c\x00\x00\x00'),
        'type' / Const(b'\x14\x00\x00\x00'),
        'status' / Const(b'\x00\x00\x00\x00'),
        'dummy' / Padding(4),
    )

    msg = fmt.build({})
    return msg


class Connection():
    """The TCP connection class."""

    def __init__(self, ps4, credential=None, port=997):
        """Init Class."""
        self.ps4 = ps4
        self._host = ps4.host
        self._credential = credential
        self._port = port
        self._socket = None
        self._cipher = None
        self._decipher = None
        self._random_seed = None
        self.pin = None

    def set_socket(self, sock):
        """Set socket."""
        self._socket = sock

    def connect(self):
        """Open the connection."""
        _LOGGER.debug('Connect')
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self._host, self._port))
        self._send_hello_request()
        data = self._recv_hello_request()
        self._set_crypto_init_vector(data.seed)
        self._send_handshake_request(data.seed)

    def disconnect(self):
        """Close the connection."""
        self._socket.close()
        self._reset_crypto_init_vector()

    def login(self, pin=None):
        """Login."""
        _LOGGER.debug('Login')
        self._send_login_request(pin=pin)
        msg = self._recv_msg()
        return _handle_response('login', msg)

    def standby(self):
        """Request standby."""
        _LOGGER.debug('Request standby')
        self._send_standby_request()
        msg = self._recv_msg()
        return _handle_response('standby', msg)

    def start_title(self, title_id):
        """Start an application/game title."""
        _LOGGER.debug('Start title: %s', title_id)
        self._send_boot_request(title_id)
        msg = self._recv_msg()
        return _handle_response('start_title', msg)

    def remote_control(self, operation, hold_time=0):
        """Send remote control command."""
        _LOGGER.debug('Remote control: %s (%s)', operation, hold_time)
        return self._send_remote_control_request(operation, hold_time)

    def send_status(self):
        """Send client connection status."""
        _LOGGER.debug('Sending Status: Connected')
        self._send_status_ack()
        msg = self._recv_msg()
        return _handle_response('send_status', msg)

    def encrypt_message(self, msg):
        """Encrypt message."""
        data = self._cipher.encrypt(msg)
        return data

    def _send_msg(self, msg, encrypted=False):
        _LOGGER.debug('TX: %s %s', len(msg), binascii.hexlify(msg))
        if encrypted:
            msg = self.encrypt_message(msg)
        self._socket.send(msg)

    def _recv_msg(self, decrypt=True):
        msg = self._socket.recv(1024)
        if decrypt:
            msg = self._decipher.decrypt(msg)
        _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        return msg

    def _set_crypto_init_vector(self, init_vector):
        self._cipher = AES.new(RANDOM_SEED, AES.MODE_CBC, init_vector)
        self._decipher = AES.new(RANDOM_SEED, AES.MODE_CBC, init_vector)

    def _reset_crypto_init_vector(self):
        self._cipher = None
        self._decipher = None

    def _send_hello_request(self):
        """Send SYN Message."""
        self._send_msg(_get_hello_request())

    def _recv_hello_request(self):
        """Receive ACK."""
        msg = self._recv_msg(decrypt=False)
        if msg is not None:
            data = _parse_hello_request(msg)
            return data
        return None

    def _send_handshake_request(self, seed):
        """Finish handshake."""
        self._send_msg(_get_handshake_request(seed))

    def _send_login_request(self, pin=None):
        name = self.ps4.device_name
        msg = _get_login_request(self._credential, name, pin)
        self._send_msg(msg, encrypted=True)

    def _send_standby_request(self):
        self._send_msg(_get_standby_request(), encrypted=True)

    def _send_boot_request(self, title_id):
        self._send_msg(_get_boot_request(title_id), encrypted=True)

    def _send_remote_control_request(self, operation, hold_time=0):
        # Prebuild required remote messages."""
        msg = _get_remote_control_request(operation, hold_time)

        for message in msg:
            try:
                self._send_msg(message, encrypted=True)
            except (socket.error, socket.timeout):
                _LOGGER.debug("Failed to send Remote MSG")

        # Delay Close RC for PS
        if operation == 128:
            _LOGGER.debug("Delaying RC off for PS Command")
            delay(1)
            try:
                self._send_msg(
                    _get_remote_control_request(operation, hold_time),
                    encrypted=True)
            except (socket.error, socket.timeout):
                _LOGGER.debug("Failed to send Remote Close MSG")

    def _send_status_ack(self):
        """Send ACK for connection status."""
        self._send_msg(_get_status_ack(), encrypted=True)


class AsyncConnection(Connection):
    """Connection using Asyncio."""

    async def async_connect(self, ps4):
        """Create asyncio TCP connection."""
        loop = asyncio.get_event_loop()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(False)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        await self._async_tcp_handshake(sock, loop)
        connect = loop.create_connection(
            lambda: TCPProtocol(ps4, loop), sock=sock)
        transport, protocol = await loop.create_task(connect)
        self.set_socket(protocol)
        return transport, protocol

    async def _async_tcp_handshake(self, sock, loop):
        await loop.sock_connect(sock, (self._host, TCP_PORT))
        msg = _get_hello_request()
        await loop.sock_sendall(sock, msg)
        response = await loop.sock_recv(sock, 1024)
        data = _parse_hello_request(response)
        self._set_crypto_init_vector(data.seed)
        msg = _get_handshake_request(data.seed)
        await loop.sock_sendall(sock, msg)


class TCPProtocol(asyncio.Protocol):
    """Asyncio TCP Protocol."""

    def __init__(self, ps4, loop):
        """Init."""
        self.loop = loop
        self.ps4 = ps4
        self.callback = _handle_response
        self.transport = None
        self.connection = ps4.connection
        self.task = None
        self.task_available = asyncio.Event()

    def connection_made(self, transport):
        """When connected."""
        self.transport = cast(asyncio.Transport, transport)
        self.ps4.connected = True
        _LOGGER.debug("PS4 Transport Connected @ %s", self.ps4.host)

        if self.ps4.task_queue is not None:
            valid_initial_tasks = {'start_title': self.start_title,
                                   "remote_control": self.remote_control}
            args = None

            initial_task = self.ps4.task_queue
            self.ps4.task_queue = None
            args = initial_task[1:]
            initial_task = initial_task[0]
            task = valid_initial_tasks.get(initial_task)
            if task is not None:
                _LOGGER.info("Queued command: %s", initial_task)
                self.task_available.set()
                if args is not None:
                    asyncio.ensure_future(task(*args))
                else:
                    asyncio.ensure_future(task())
        else:
            self.task_available.set()

    def data_received(self, data):
        """Call when data received."""
        data = self.connection._decipher.decrypt(data)  # noqa: pylint: disable=protected-access
        self._handle(data)

    def connection_lost(self, exc):
        """Call if connection lost."""
        self.disconnect()

    def _complete_task(self):
        self.task = None
        self.task_available.set()

    async def add_task(self, task_name, func, *args):
        """Add task to queue."""
        if args:
            task = func(*args)
        else:
            task = func()

        await self.task_available.wait()
        self.task_available.clear()
        self.task = task_name
        await task

    async def send(self, msg):
        """Send Message."""
        _LOGGER.debug('TX: %s %s', len(msg), binascii.hexlify(msg))
        msg = self.connection.encrypt_message(msg)
        self.transport.write(msg)

    def sync_send(self, msg):
        """Send Message synchronously."""
        _LOGGER.debug('TX: %s %s', len(msg), binascii.hexlify(msg))
        msg = self.connection.encrypt_message(msg)
        self.transport.write(msg)

    def _handle(self, data):
        data_hex = binascii.hexlify(data)
        _LOGGER.debug('RX: %s %s', len(data), data_hex)
        if data_hex == STATUS_REQUEST:
            self._ack_status()
        elif self.task == 'remote_control':
            pass
        else:
            if self.callback(self.task, data):
                _LOGGER.debug("Command successful: %s", self.task)
                if self.task == 'login':
                    self.ps4.loggedin = True
                self._complete_task()

    async def login(self, pin=None):
        """Login."""
        task_name = 'login'
        name = self.ps4.device_name
        msg = _get_login_request(self.ps4.credential, name, pin)
        task = self.add_task(task_name, self.send, msg)
        asyncio.ensure_future(task)

    async def standby(self):
        """Standby."""
        if not self.ps4.loggedin:
            await self.login()
        task_name = 'standby'
        msg = _get_standby_request()
        task = self.add_task(task_name, self.send, msg)
        asyncio.ensure_future(task)

    def disconnect(self):
        """Close the connection."""
        if self.transport is not None:
            self.transport.close()
            self.ps4.protocol = None
            self.transport = None
            _LOGGER.debug("Transport @ %s is disconnected", self.ps4.host)
        self.connection._reset_crypto_init_vector()  # noqa: pylint: disable=protected-access
        self.ps4.loggedin = False
        self.ps4.connected = False

    async def start_title(self, title_id, running_id=None):
        """Start Title."""
        if not self.ps4.loggedin:
            await self.login()
        task_name = 'start_title'
        msg = _get_boot_request(title_id)
        task = self.add_task(task_name, self.send, msg)
        asyncio.ensure_future(task)
        if running_id != title_id:
            self.loop.call_later(1.0, self._send_remote_control_request, 16)

    async def remote_control(self, operation, hold_time=0):
        """Remote Control."""
        if not self.ps4.loggedin:
            await self.login()
        task_name = 'remote_control'
        msg = _get_remote_control_request(operation, hold_time)
        task = self.add_task(task_name, self._send_remote_control_request,
                             msg, operation)
        asyncio.ensure_future(task)

    async def _send_remote_control_request(self, msg, operation):
        """Send messages for remote control."""
        for message in msg:
            # Messages are time sensitive.
            # Needs to be immediately sent in order.
            self.sync_send(message)

        # For 'PS' command.
        if operation == 128:
            # Even more time sensitive. Delay of 1 Second needed.
            self.loop.call_later(
                1.0, self.sync_send, _get_remote_control_close_request())

        # Don't handle or wait for a response
        self.task_available.set()

    def _ack_status(self):
        """Sends msg in response to heartbeat message."""
        # Update state as well, no need to manage polling now.
        self.ps4.get_status()
        asyncio.ensure_future(self.send(_get_status_ack()))
        _LOGGER.debug("Sending Hearbeat response")



# -*- coding: utf-8 -*-
"""Device Discovery Protocol for PS4."""
from __future__ import print_function

import re
import socket
import logging
import select
import asyncio

_LOGGER = logging.getLogger(__name__)

BROADCAST_IP = '255.255.255.255'
UDP_IP = '0.0.0.0'
UDP_PORT = 0

DDP_PORT = 987
DDP_VERSION = '00020020'

DEFAULT_POLL_COUNT = 3


class DDPProtocol(asyncio.DatagramProtocol):
    """Async UDP Client."""

    def __init__(self, max_polls=DEFAULT_POLL_COUNT):
        """Init Instance."""
        self.callbacks = {}
        self.transport = None
        self.port = DDP_PORT
        self.max_polls = max_polls
        self.message = get_ddp_search_message()
        super().__init__()

    def _set_write_port(self, port):
        """Only used for tests."""
        self.port = port

    def connection_made(self, transport):
        """On Connection."""
        self.transport = transport
        _LOGGER.debug("PS4 Transport Created: %s", type(self.transport))

    def send_msg(self, ps4, message=None):
        """Send Message."""
        if message is None:
            message = self.message

        self.transport.sendto(message.encode('utf-8'),
                              (ps4.host, self.port))

        # Track polls that were never returned.
        ps4.poll_count += 1

        # Assume PS4 is not available.
        if ps4.poll_count >= self.max_polls:
            if not ps4.unreachable:
                _LOGGER.info("PS4 @ %s is unreachable", ps4.host)
                ps4.unreachable = True
                ps4._close()  # noqa: pylint: disable=protected-access
            ps4.status = None
            if ps4.host in self.callbacks:
                callback = self.callbacks[ps4.host].get(ps4)
                if callback is not None:
                    callback()

    def datagram_received(self, data, addr):
        """When data is received."""
        if data is not None:
            self._handle(data, addr)

    def _handle(self, data, addr):
        data = parse_ddp_response(data.decode('utf-8'))
        data[u'host-ip'] = addr[0]

        address = addr[0]

        if address in self.callbacks:
            for ps4, callback in self.callbacks[address].items():
                ps4.poll_count = 0
                ps4.unreachable = False
                old_status = ps4.status
                ps4.status = data
                if old_status != data:
                    _LOGGER.debug("Status: %s", ps4.status)
                    callback()

    def connection_lost(self, exc):
        """On Connection Lost."""
        if self.transport is not None:
            _LOGGER.error("DDP Transport Closed")
            self.transport.close()

    def error_received(self, exc):
        """Handle Exceptions."""
        _LOGGER.warning("Error received at DDP Transport")

    def close(self):
        """Close Transport."""
        self.transport.close()
        self.transport = None
        _LOGGER.info("Closing DDP Transport")

    def add_callback(self, ps4, callback):
        """Add callback to list. One per PS4 Object."""
        if ps4.host not in self.callbacks:
            self.callbacks[ps4.host] = {}
        self.callbacks[ps4.host][ps4] = callback

    def remove_callback(self, ps4, callback):
        """Remove callback from list."""
        if ps4.host in self.callbacks:
            if self.callbacks[ps4.host][ps4] == callback:
                self.callbacks[ps4.host].pop(ps4)

                # If no callbacks remove host key also.
                if not self.callbacks[ps4.host]:
                    self.callbacks.pop(ps4.host)


async def async_create_ddp_endpoint():
    """Create Async UDP endpoint."""
    loop = asyncio.get_event_loop()
    connect = loop.create_datagram_endpoint(
        lambda: DDPProtocol(), local_addr=(UDP_IP, UDP_PORT),  # noqa: pylint: disable=unnecessary-lambda
        reuse_address=True, reuse_port=True, allow_broadcast=True)
    transport, protocol = await loop.create_task(connect)
    return transport, protocol


def get_ddp_message(msg_type, data=None):
    """Get DDP message."""
    msg = u'{} * HTTP/1.1\n'.format(msg_type)
    if data:
        for key, value in data.items():
            msg += '{}:{}\n'.format(key, value)
    msg += 'device-discovery-protocol-version:{}\n'.format(DDP_VERSION)
    return msg


def parse_ddp_response(rsp):
    """Parse the response."""
    data = {}
    app_name = None
    for line in rsp.splitlines():
        if 'running-app-name' in line:
            app_name = line
            app_name = app_name.replace('running-app-name:', '')
        re_status = re.compile(r'HTTP/1.1 (?P<code>\d+) (?P<status>.*)')
        line = line.strip()
        # skip empty lines
        if not line:
            continue
        elif re_status.match(line):
            data[u'status_code'] = int(re_status.match(line).group('code'))
            data[u'status'] = re_status.match(line).group('status')
        else:
            values = line.split(':')
            data[values[0]] = values[1]
    if app_name is not None:
        data['running-app-name'] = app_name
    return data


def get_ddp_search_message():
    """Get DDP search message."""
    return get_ddp_message('SRCH')


def get_ddp_wake_message(credential):
    """Get DDP wake message."""
    data = {
        'user-credential': credential,
        'client-type': 'a',
        'auth-type': 'C',
    }
    return get_ddp_message('WAKEUP', data)


def get_ddp_launch_message(credential):
    """Get DDP launch message."""
    data = {
        'user-credential': credential,
        'client-type': 'a',
        'auth-type': 'C',
    }
    return get_ddp_message('LAUNCH', data)


def get_socket(timeout=3):
    """Return DDP socket object."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(timeout)
    sock.bind((UDP_IP, UDP_PORT))
    return sock


def _send_recv_msg(host, broadcast, msg, receive=True):
    """Send a ddp message and receive the response."""
    sock = get_socket()

    if broadcast:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        _host = host or BROADCAST_IP
    else:
        _host = host

    sock.sendto(msg.encode('utf-8'), (_host, DDP_PORT))

    if receive:
        available, _, _ = select.select([sock], [], [], 0.01)
        if sock in available:
            return sock.recvfrom(1024)
    return None


def _send_msg(host, broadcast, msg):
    """Send a ddp message."""
    return _send_recv_msg(host, broadcast, msg, receive=False)


def send_search_msg(host):
    """Send message only."""
    msg = get_ddp_search_message()
    return _send_msg(host, True, msg)


def search(host=None, broadcast=True):
    """Discover PS4s."""
    msg = get_ddp_search_message()
    data, addr = _send_recv_msg(host, broadcast, msg)
    if data is not None:
        ps_list = []
        data = parse_ddp_response(data.decode('utf-8'))
        data[u'host-ip'] = addr[0]
        ps_list.append(data)
        return ps_list
    return None


def get_status(host):
    """Get status."""
    try:
        ps_list = search(host=host)
    except TypeError:
        return None
    return ps_list[0]


def wakeup(host, credential, broadcast=None):
    """Wakeup PS4s."""
    msg = get_ddp_wake_message(credential)
    _send_msg(host, broadcast, msg)


def launch(host, credential, broadcast=None):
    """Launch."""
    msg = get_ddp_launch_message(credential)
    _send_msg(host, broadcast, msg)


class Discovery:
    """Device Discovery server."""

    def __init__(self):
        """Init."""
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(6.0)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.msg = get_ddp_search_message()
        self.host = '255.255.255.255'
        self.ps_list = []

    def search(self, host):
        """Search for Devices."""
        if host is None:
            host = self.host
        finished = False
        try:
            self.send(host)
        except (socket.error, socket.timeout):
            self.sock.close()
            return self.ps_list
        while finished is False:
            try:
                device = self.receive()
                if device not in self.ps_list:
                    self.ps_list.append(device)
            except (socket.error, socket.timeout):
                self.sock.close()
                self.ps_list = []
            return self.ps_list

    def send(self, host):
        """Broadcast Message."""
        self.sock.sendto(self.msg.encode('utf-8'), (host, DDP_PORT))

    def receive(self):
        """Receive Message."""
        data, addr = self.sock.recvfrom(1024)
        if data is not None:
            data = parse_ddp_response(data.decode('utf-8'))
            data[u'host-ip'] = addr[0]
            return data
        return None
